import World
import Route 


class Move:
    def __init__(self, bid, move):
        self._bid = bid
        self._move = move


class Robot:
    def __init__(self, id):
        self.id = id
        self._x, self._y = 0, 0
        self._packages = []
        print "New robot:", id

    def move_by(self, x, y):
        self._x += x
        self._y += y
        self._update_pkg_positions()

    def set_position(self, x, y):
        self._x = x
        self._y = y
        self._update_pkg_positions()

    def position(self):
        return (self._x, self._y)

    def can_carry(self, pkg):
        return pkg.weight + self._load < self._capacity

    def pick_pkg(self, pkg):
        if pkg:
            self._packages.append(pkg)
            pkg.set_position(self._x, self._y)
            pkg.pick()

    def drop_pkg(self, pkg):
        if pkg:
            try:
                pkg.set_position(self._x, self._y)
                pkg.drop()
                self._packages.remove(pkg)
            except ValueError:
                pass

    def _update_pkg_positions(self):
        for package in self._packages:
            package.set_position(self._x, self._y)

    def packages(self):
        return self._packages

    def die(self):
        self._packages = []

# ###########################################

class Player(Robot):
    def __init__(self, id, capacity, money):
        Robot.__init__(self, id) 
        self._capacity = capacity
        self._money = money
        self._load = 0
        self._world = World.get_world()
        self._route = []
        self._wander = [] 
        print 'I am %d carry %d, at (%d,%d)' % (self.id, self._capacity, self._x, self._y)

    def _determine_route(self):
        delivery_threshold = (self._capacity / 10) * 5  # 50%
        # Determine shortest Drop route
        drop_route = self.get_next_drop()
        print "Load:", self._load
        if self._load > delivery_threshold:
            self._route = drop_route
        else:
            # Determine shortest Pick route
            print "Bases", len(self._world._bases)
            pick_route = self._world.get_next_base(self._x, self._y)

            if not drop_route and not pick_route:
                print "Packages", len(self._world._packages.keys())
                self._route = self._world.get_next_package(self._x, self._y)
            else:
                if drop_route:
                    if not pick_route or len(drop_route) <= len(pick_route):
                        self._route = drop_route
                    else:
                        self._route = pick_route
                else:  # must have a pick route
                    self._route = pick_route

    def calculate_move(self):
        # Calculate risk from other players
        bid_amt = self._guess_bid()

        # Drop Packages
        drop_pkgs = [ p for p in self._packages
                      if p.destination() == (self._x, self._y) ]
        param = self._pkg_list(drop_pkgs)
        if param:
            bid = self._decr_money(bid_amt)
            action = 'Drop'
            return self._cmd_str(bid, action, param)

        # Pick Packages
        pick_pkgs = self.calculate_load()
        if pick_pkgs:
##            print "Pick Pkgs:",
##            for p in pick_pkgs:
##                print p.id,
##            print
            param = self._pkg_list(pick_pkgs)
            if param:
                bid = self._decr_money(bid_amt)
                action = 'Pick'
                return self._cmd_str(bid, action, param)

        if not self._route:
            # We can't be wandering if no route 
            if self._wander:
                self._wander = []
            self._determine_route()
        elif self._wander:
            # Why are we wandering? Lets stop
            print "Wandering..."
            action = 'Move'
            x,y = self._wander[-1]
            bid = self._decr_money(bid_amt)
            param = self.direction_of(x,y)
            return self._cmd_str(bid, action, param)

        if self._route:
            # We should be on our path  
            action = 'Move'
            x,y = self._route[0]
            bid = self._decr_money(bid_amt)
            param = self.direction_of(x,y)
            return self._cmd_str(bid, action, param)

        print 'No more paths to follow, aborting.'
        return ''

    def _cmd_str(self, bid, action, param):
        return '%d %s %s' % (bid, action, param)

    def _decr_money(self, bid):
        self._money -= abs(bid)
        return bid

    def _pkg_list(self, packages):
        tmp_packages = []
        for package in packages:
            tmp_packages.append(package.id)
        return ' '.join(map(str, tmp_packages))

    def _guess_bid(self):
        n,s,e,w = self._world.get_adjacent(self._x, self._y)
        bid = 1 
                
        if not n : bid += 1                
        if not s : bid += 1                
        if not e : bid += 1                 
        if not w : bid += 1                 
                    
        if (not n and s == World.Water) or (not s and n == World.Water):
            bid = bid * 3
        if (not e and w == World.Water) or (not w and e == World.Water):
            bid = bid * 3

        return bid

    def direction_of(self, x, y):
        if self._y < y: return 'N'
        if self._y > y: return 'S'
        if self._x < x: return 'E'
        if self._x > x: return 'W'

    def move_by(self, x, y):
        ox, oy = self._x, self._y
        Robot.move_by(self,x,y)
        self.update(ox, oy)

    def set_position(self, x, y):
        ox, oy = self._x, self._y
        Robot.set_position(self,x,y)
        if (ox, oy) != (self._x, self._y):
            self.update(ox, oy)

    def get_next_drop(self):
        print "Drops:",
        routes = []
        dests = []
        for package in self._packages:
            dx, dy = package.destination()
            if (dx, dy) not in dests:
                dests.append((dx, dy))
                route = Route.Route(dx, dy)
                path = route.find_path(self._x, self._y)
                routes.append((len(path), path))
        routes.sort()
        print len(routes)
        if routes:
            return routes[0][1]
        return []

    def update(self, ox, oy):
        if self._route:
            if (self._x,self._y) == self._route[0]:
                self._route.pop(0)
            else:
                if self._wander and (self._x,self._y) == self._wander[-1]:
                    self._wander.pop()
                else:
                    self._wander.append((ox, oy))

    def pick_pkg(self, pkg):
        if pkg:
            self._packages.append(pkg)
            self._load += pkg.weight()
            pkg.set_position(self._x, self._y)
            pkg.pick()

    def drop_pkg(self, pkg):
        if pkg:
            try:
                self._load -= pkg.weight()
                pkg.set_position(self._x, self._y)
                pkg.drop()
                if (self._x, self._y) == pkg.destination():
                    self._world.remove_package(pkg)
                self._packages.remove(pkg)
            except ValueError:
                pass

    def calculate_load(self):
        best_pkgs = []
        # Collect all packages we're carrying plus all packages at this square.
        world_packages = self._world.packages_at(self._x, self._y)
##        print "World Pkgs:",
##        for p in world_packages:
##            print p.id,
##        print
        world_packages = [ p for p in world_packages
                           if p not in self._packages ]
##        print "World Pkgs not in Player._packages:",
##        for p in world_packages:
##            print p.id,
##        print
        if world_packages:
##            avail_packages = self._packages[:] + world_packages
            avail_packages = world_packages
            # Calculate the best arrangement in our cargo hold.
            best_pkgs = self._best_load2(avail_packages, self._capacity - self._load)
        return best_pkgs
  
    def _best_load(self, items, capacity):
        P = {}
        for i in xrange(len(items)+1):
            for lim in xrange(capacity+1):
                if i == 0:
                    P[i,lim] = 0
                elif items[i-1].weight() > lim:
                    P[i,lim] = P[i-1,lim]
                else:
                    P[i,lim] = max(P[i-1,lim],
                        P[i-1,lim-items[i-1].weight()] +
                            items[i-1].weight())

        L = []
        i = len(items)
        lim = capacity
        while i > 0:
            if P[i,lim] == P[i-1,lim]:
                i -= 1
            else:
                i -= 1
                L.append(items[i])
                lim -= items[i].weight()

        L.reverse()
        return L

    def _best_load2(self, items, capacity):
        P = {}

        def recurse(i, lim):
            if not P.has_key((i, lim)):
                if i == 0:
                    P[i, lim] = 0
                elif items[i-1].weight() > lim:
                    P[i, lim] = recurse(i-1, lim)
                else:
                    P[i, lim] = max(recurse(i-1, lim),
                        recurse(i-1,
                                lim-items[i-1].weight()) +
                            items[i-1].weight())
            return P[i, lim]

        recurse(len(items), capacity)
        L = []
        i = len(items)
        lim = capacity
        while i > 0:
            if P[i,lim] == P[i-1,lim]:
                i -= 1
            else:
                i -= 1
                L.append(items[i])
                lim -= items[i].weight()

        L.reverse()
        return L

    def die(self):
      print 'Died with ID:', self.id, 'at', (self._x,self._y),
      print 'and', self._money, 'fuel remaining.'
      print 'Current Packages:',
      for package in self._packages:
          print package.id,
      print
