import World


N = 0
S = 1
E = 2
W = 3

_coord_funcs = [ lambda x, y: (x, y + World.North),
                 lambda x, y: (x, y + World.South),
                 lambda x, y: (x + World.East, y),
                 lambda x, y: (x + World.West, y) ]


class Route:
  def __init__(self, dx, dy):
    self.dx = dx
    self.dy = dy
    self._path = []
    self._visited = {}
    self.world = World.get_world()

  def coords(self, dir, x, y):
    return _coord_funcs[dir](x, y)

  def move(self, dir, x, y, block=0):
    if dir == N: y+=World.North
    if dir == S: y+=World.South
    if dir == W: x+=World.West 
    if dir == E: x+=World.East
    self._path.append((x, y))
    self._visited[(x, y)] = None
    return (x, y)

  def has_choices(self, x, y):
    for dir in range(4):
      if self.is_clear(dir, x, y) and not self.visited(dir, x, y):
        return 1
    return 0

  def backup(self):
    tx, ty = self._path[-1]
    if self.has_choices(tx, ty):
      return (tx, ty)
    self._path.pop()
    return self.backup()

  def is_clear(self, dir, x, y):
    px, py = self.coords(dir, x, y)
    object = self.world.whats_at(px, py)
    if object == World.Clear or object == World.Base:
      return 1
    return 0

  def visited(self, dir, x, y):
    return self._visited.has_key(self.coords(dir, x, y))

  def find_path(self, x, y):
#    self._path.append((x, y))
    while x != self.dx or y != self.dy:
      # Go somewhere sensible
      if x > self.dx and self.is_clear(W, x, y) and not self.visited(W, x, y): # Go West
#        print 'West looks good'
        x, y = self.move(W, x, y) 
      elif x < self.dx and self.is_clear(E, x, y) and not self.visited(E, x, y): # Go East
#        print 'East looks good'
        x, y = self.move(E, x, y) 
      elif y > self.dy and self.is_clear(S, x, y) and not self.visited(S, x, y): # Go South
#        print 'South looks good'
        x, y = self.move(S, x, y)
      elif y < self.dy and self.is_clear(N, x, y) and not self.visited(N, x, y): # Go North
#        print 'North looks good'
        x, y = self.move(N, x, y)

      # Try to move somewhere new atleast
      elif self.is_clear(W, x, y) and not self.visited(W, x, y): # Go West
#        print 'West is new'
        x, y = self.move(W, x, y)
      elif self.is_clear(E, x, y) and not self.visited(E, x, y): # Go East
#        print 'East is new'
        x, y = self.move(E, x, y)
      elif self.is_clear(S, x, y) and not self.visited(S, x, y): # Go South
#        print 'South is new'
        x, y = self.move(S, x, y)
      elif self.is_clear(N, x, y) and not self.visited(N, x, y): # Go North
#        print 'North is new'
        x, y = self.move(N, x, y)

      # Get outta there
      else:
#        print 'Backup'
        x, y = self.backup()

##    orig_len = len(self._path)
    self._path = self.optimize(self._path)
##    print 'Found:', (self.dx, self.dy),
##    print 'in %d, now %d moves' % (orig_len, len(self._path))
##    self.verify_path()

    return self._path

  def optimize(self, path):
    i = 0
    while i < len(path) - 3:
      cur_x, cur_y = path[i]
      j = len(path) - 1
      while j > i + 2:
        tst_x, tst_y = path[j]
        d = abs(cur_x - tst_x) + abs(cur_y - tst_y)
        if d == 1:
          path = path[:i+1] + path[j:]
          break
        j -= 1
      i += 1
    return path

  def verify_path(self):
    print "Verifying:",
    bad_coords = []
    i = 0
    while i < len(self._path) - 1:
      cur_x, cur_y = self._path[i]
      tst_x, tst_y = self._path[i+1]
      d = abs(cur_x - tst_x) + abs(cur_y - tst_y)
      if d != 1:
        bad_coords.append((i, i + 1))
      i += 1
    print "done.",

    if bad_coords:
      for i in range(len(self._path)):
        if i % 6 == 0:
          print
        print "(%2d, %2d) " % self._path[i],

      print "\nBad:",
      for t in bad_coords:
        print t,
    print

